

#include "CGameControl.h"
#include "ITime.h"
#include "CPlayer.h"
#include "ILevelLoader.h"
#include "stateflagdefs.h"
#include "globalconst.h"
#include "Vec3.h"
#include "soundfunc.h"
#include "IInput.h"
#include "IHighScores.h"
#include "filedefs.h"
#include "ITextToSpeech.h"
#include <string.h>
#include <iostream>

// -------------------------------------
// CGameControl - functions definitions
//
// ... see header for more information
//
//	UPDATES:
//		- 2/7/04, Daniel: created
//		- 2/9/04, Daniel: CoinPickedUp, commented stuff
//		- 2/11/04, Daniel: racethink
//		- 2/12/04, Daniel: CalcScore, more to racethink, CheckCollision
//		- 2/17/04, Daniel: The level loading system now works, LoadLevel
//		- 2/18/04, Daniel: debug text graphics
//		- 2/19/04, Daniel: now handling level endings
//		- 2/22/04, Daniel: PlaySounds
//		- 2/24/04, Patrick: Draw changes
//		- 3/1/04,  Daniel: now tracks player's lives
//		- 3/21/04, Daniel: call to SwitchCheck
//		- 3/22/04, Daniel: SoundFunc stuff
//		- 3/31/04, Daniel: WarningThink
//		- 4/13/04, Daniel: PauseThink
//		- 4/17/04, Daniel: HighScoresThink, LevelSuccessThink
//
// -------------------------------------



//
//	Default constructor 
//			
//
CGameControl :: CGameControl() 
{
	entCount = -1; // array is not ready
	gameState = GAME_INIT; // set state to load level
	level = 1; // set level to the first
	score = 0;
	coins = 0;
	entArray = NULL; // initialize
	m_pHighScores = NULL;
	loader = new ILevelLoader();
	pauseTime = 0;
	pauseBeginTime = 0;
	levelDoneTime = 0;
	levelEndTime = 0;
	levelStartTime = 0;
	lives = NUM_LIVES_STANDARD;
	levelLength = 0;
	prePauseState = -1;

	name[0] = 'A';
	name[1] = 'A';
	name[2] = 'A';
	timerLetterRemind = 0.0f;
	nameState = 0;
	timerConfirm = 0.0f;
	timerBack = 0.0f;
	timerUp = 0.0f;
	timerDown = 0.0f;
	timerRight = 0.0f;
	timerLeft = 0.0f;

	bLevelEnd = false;
	timeTillSayScores = 0.0f;
	timeTillCanContinue = 0.0f;

	timerDeathPlayCredits = 0.0f;
	timerDeathTransition = 0.0f;
}

//
//	Level start constructor
//		
//
CGameControl :: CGameControl( int startLevel )
{
	entCount = -1; // array is not ready 
	gameState = GAME_INIT; // set state to load level
	level = startLevel; // set level to given value
	score = 0;
	coins = 0;
	entArray = NULL; // initialie to null
	loader = new ILevelLoader();
	m_pHighScores = NULL;
	pauseTime = 0;
	pauseBeginTime = 0;
	levelDoneTime = 0;
	levelEndTime = 0;
	levelStartTime = 0;
	lives = NUM_LIVES_STANDARD;
	levelLength = 0;
	prePauseState = -1;

	name[0] = 'A';
	name[1] = 'A';
	name[2] = 'A';
	timerLetterRemind = 0.0f;
	nameState = 0;
	timerConfirm = 0.0f;
	timerBack = 0.0f;
	timerUp = 0.0f;
	timerDown = 0.0f;
	timerRight = 0.0f;
	timerLeft = 0.0f;

	bLevelEnd = false;
	timeTillSayScores = 0.0f;
	timeTillCanContinue = 0.0f;

	timerDeathPlayCredits = 0.0f;
	timerDeathTransition = 0.0f;
}

//
// CleanupEnts
//	... called by: LevelSuccessThink/DeathThink
//	... description: free our array of CBaseEntitys
//
void CGameControl :: CleanupEnts()
{
	for ( int i = 0; i < entCount; i++ )
		delete entArray[i];
	delete entArray;

	entArray = NULL;
}

//
// CleanupEnts
//	... called by: DeathThink
//	... description: make ILevelLoader free its database of segments
//
void CGameControl :: CleanupLoader()
{
	loader->Cleanup();
	delete loader;

	loader = NULL;
}

//
// Think
//	... called by: CUberControl::Think()
//	... description: frame-by-frame logic
//
void CGameControl :: Think()
{

	// Get an action depending on what state we're in
	switch ( gameState )
	{
		case GAME_INIT:			// used: once, when CGameControl is created
			LevelLoadData();	// have ILevelLoader create a database from the level file
			IInitializeRaceSounds( entArray, entCount );	// load the sound samples for use in the game
			break;
		case GAME_LEVELINIT:	// used: once at the beginning of each level
			SpeakLevelInfo();	// have text to speech say what level we're playing
			LevelGenerate();	// have ILevelLoader create an array of CBaseEntitys from its segment database
			pauseTime = 0;		// reset timers
			pauseBeginTime = 0;	// reset timers
			IResetRaceSounds(entArray,entCount);	// reset/initialize our entities
			IPlayStationarySound(NULL,40,SOUND_EVENT_LEVELSTART); // start playing the level start sound
			break;
		case GAME_STARTING:		// used: every frame during the starting sequence at the beginning of every level
			StartThink();		
			break;
		case GAME_RACING:		// used: every frame while the player is racing down the track
			RaceThink();
			break;
		case GAME_PAUSE:		// used: every frame while paused
			PauseThink();
			break;
		case GAME_LEVEL_SUCCESS:	// used: every frame after the level has ended, before the next one starts
			LevelSuccessThink();
			break;
		case GAME_HIGHSCORES:	// used: every frame while the player is inputting a high score
			HighScoresThink();
			break;
		case GAME_DEATH:		// used: every frame while the player is dead, but hasn't returned to the main menu
			DeathThink();
			break;
		case GAME_QUIT:			// used: dead-end state, must be picked up by the outside as a signal of our intent to quit
			break;

	}
}


//
// StartThink
//  ... called by: Think
//	... description: think code while in the starting racing mode
//
void CGameControl :: StartThink()
{
	// are we asking to pause?
	if ( glInput->isKeyPressed( GKEY_ESC ) )
	{
		pauseBeginTime = glTime->getTime(); // this is when we paused
		gameState = GAME_PAUSE;	// switch to pause mode
		IPlayStationarySound(NULL,50,SOUND_EVENT_PAUSE); // play the paused sound
		IPauseSounds(); // pause all the sounds
		prePauseState = GAME_STARTING; // when we paused, we were in this state
		return;
	}

	// is it time to start yet?
	if ( glTime->getTime() >=  timeTillStart )
	{
		gameState = GAME_RACING; // switch to racing
		levelStartTime = glTime->getTime(); // we started now
		pauseTime = 0;		// reset pauseTime
		pauseBeginTime = 0; // reset the pauseBeginTime
	}

	// update our pre-game camera
	for ( int i = 0; i < entCount; i++ )
	{
		if ( entArray[i]->pvars->obj_type == OBJ_CAMERA )
		{
			if ( entArray[i]->pvars->thinktime < glTime->getTime() )
				entArray[i]->Think(); // CAMERA THINK
		}
	}
}


//
// RaceThink
//  ... called by: Think
//	... description: think code while in normal racing mode
//
void CGameControl :: RaceThink()
{
	if ( glInput->isKeyPressed( GKEY_ESC ) )
	{
		// escape has been pressed, go into pause mode
		pauseBeginTime = glTime->getTime(); // we just started the pause
		gameState = GAME_PAUSE; // change our state
		IPauseSounds();		// don't hear the object sounds while paused
		IPlayStationarySound(NULL,50,SOUND_EVENT_PAUSE);	// play the pause sound
		prePauseState = GAME_RACING;	// we were racing before we paused
		return;
	}

	// cycle through all CBaseEntities in entArray
	for ( int i = 0; i < entCount; i++ )
	{
		CBaseEntity *pEnt = entArray[i];

		if ( pEnt->pvars->ent_flags & ENT_MOVING )
			CheckCollisions( pEnt ); // check for collisions

		if ( pEnt->isPlayer() )
		{
			((CPlayer*)pEnt)->HandleInput(); // do input for the player
			((CPlayer*)pEnt)->SwitchCheck(); // if we're switching, change our x value
		}

		// think at intervals
		if ( pEnt->pvars->thinktime < glTime->getTime() )
		{
			pEnt->Think(); // MAIN THINK CALL

			// if we're the player, do some additional checks
			if ( pEnt->isPlayer() )
			{
				// get player stats update
				CPlayer *pPlayer = (CPlayer*)pEnt;
				coins = pPlayer->coins; // update our coins
				lives = pPlayer->lives;	// update our lives
				pPlayer->level = level;	// tell the player what level we're in

				if ( pPlayer->pvars->origin.z > levelLength ) // we're at the end of the level
				{
					IPlayStationarySound(NULL, 30, SOUND_EVENT_LEVELEND ); // play the level end sound
					gameState = GAME_LEVEL_SUCCESS; // switch to level success mode
					levelEndTime = glTime->getTime();	// this is the time we ended on
					levelDoneTime = (levelEndTime - levelStartTime) - pauseTime; // calculate the total time spent
					CalcScore();	// calculate score
					coins = 0; // reset the coin values

				}
				if ( pPlayer->pvars->player_state == PL_DEAD )
				{
					IPauseSounds();	// we're dead, pause sounds
					IPlayStationarySound(NULL, 100, SOUND_EVENT_DEATH ); //play the death sound
					gameState = GAME_DEATH; // game is over
					levelEndTime = glTime->getTime(); // this is when the level has ended
					levelDoneTime = (levelEndTime - levelStartTime) - pauseTime; // end-start, minus paused time
					CalcScore(); // give the player credit for coins, not bonus time.
					coins = 0;	// we forfeit our coins

					// check to see if we had a high score when we died
					if ( !m_pHighScores )
						m_pHighScores = new IHighScores( FILE_HIGHSCORES );
					if ( m_pHighScores->checkScore( score ) ) 
					{
						// yes, we did
						gameState = GAME_HIGHSCORES; // switch to high scores mode
						IPauseSounds();	// pause the sounds (redundant?)
						IPlayStationarySound(NULL,100,SOUND_EVENT_FIRSTINITIAL); // play the first initial sound... this should probably be more elegant
					}

				}
			}
		}
	}
}

//
// PauseThink
//  ... called by: Think
//	... description: think code while paused
//
void CGameControl :: PauseThink()
{
	if ( glTime->getTime() - pauseBeginTime > 0.5f )
	{
		if ( glInput->isKeyPressed( GKEY_ESC ) )
		{
			if ( !m_pHighScores )
				m_pHighScores = new IHighScores( FILE_HIGHSCORES ); // create an IHighScores object
			if ( m_pHighScores->checkScore( score ) )
			{
				// we're quitting and we have a high score
				gameState = GAME_HIGHSCORES;
				IPauseSounds();	// don't hear any of the objects while doing high scores
				timerBack = glTime->getTime() + GAME_TIMER_BACK;	// so that we don't exit out immediately
				IPlayStationarySound(NULL,100,SOUND_EVENT_FIRSTINITIAL); // we start on the first initial
				return;
			}

			// we're quitting and we do not have a high score
			gameState = GAME_DEATH;	// go to death mode
			levelEndTime = pauseBeginTime; // the level ended when we paused, retroactively
			timerDeathTransition = glTime->getTime() + GAME_TIMER_DEATHTRANS; // so we don't exit out immediately
			timerDeathPlayCredits = timerDeathTransition;	// start playing credits when the transition is over
			IPlayStationarySound(NULL,100,SOUND_EVENT_DEATH);	// play the death sound (roughly as long as the transition)
			return;
		}
	}
	if ( glInput->isKeyPressed( GKEY_SPACE ) )
	{
		// we're going back into the game
		gameState = prePauseState;	// go back to the state we were in previously
		pauseTime += glTime->getTime() - pauseBeginTime; // increase the amount of time we need to subtract from the final time
		pauseBeginTime = 0; // reset this time value
		IUnpauseSounds(); // start the sounds again
		return;
	}
}

//
// HighScoresThink
//  ... called by: Think
//	... description: think code while processing high scores
//
void CGameControl :: HighScoresThink()
{
	if ( glInput->isKeyPressed( GKEY_ESC ) && timerBack <= glTime->getTime() )
	{
		// we've pressed escape, and are going backwards
		if ( nameState == 0 )
		{
			// we're at the first initial and going back
			m_pHighScores = NULL;
			gameState = GAME_DEATH; // switch to death mode
			timerDeathTransition = glTime->getTime() + GAME_TIMER_DEATHTRANS; // so we don't exit out immediately
			timerDeathPlayCredits = timerDeathTransition; // start credits when the transition ends
			IPlayStationarySound(NULL,100,SOUND_EVENT_DEATH); // transition sound
			return;
		}
		else if ( nameState == 1 )
		{
			nameState = 0; // go back to the first initial
			IPlayStationarySound(NULL,100,SOUND_EVENT_FIRSTINITIAL); // "First Initial"
		}
		else if ( nameState == 2 )
		{
			nameState = 1; // skip to the second initial
			IPlayStationarySound(NULL,100,SOUND_EVENT_SECONDINITIAL); // "Second Initial"
		}

		timerBack = glTime->getTime() + GAME_TIMER_BACK;
	}

	if ( glInput->isKeyPressed( GKEY_SPACE ) && timerConfirm <= glTime->getTime() )
	{
		// we've pressed space, and are going forwards in the progression from quitting to entering out initials
		if ( nameState == 0 )
		{
			IPlayStationarySound(NULL,100,SOUND_EVENT_SECONDINITIAL); // we're on the "Second initial"
			nameState = 1; // go to the next state, 1
		}
		else if ( nameState == 1 )
		{
			IPlayStationarySound(NULL,100,SOUND_EVENT_THIRDINITIAL); // we're on the "Third initial"
			nameState = 2;
		}
		else if ( nameState == 2 )
		{
			string s = "";
			// we're adding our high score to the list...
			// create a string int the form "A.A.A."
			char c[6];
			c[0] = name[0];
			c[1] = '.';
			c[2] = name[1];
			c[3] = '.';
			c[4] = name[2];
			c[5] = '.';
			s.append(c,6);

			// input our score and our name into m_pHighScores
			m_pHighScores->checkScore( s, score ); // this also involves a substantial amount of TTS
			m_pHighScores = NULL;	// reset the pointer TODO: should be delete?
			gameState = GAME_DEATH; // switch to death mode
			timerDeathTransition = glTime->getTime() + GAME_TIMER_DEATHTRANS; // so we don't exit out immediately
			timerDeathPlayCredits = timerDeathTransition; // play the credits when the transition ends
			IPlayStationarySound(NULL,100,SOUND_EVENT_DEATH);
		}
	
		timerConfirm = glTime->getTime() + GAME_TIMER_CONFIRM;
	}
	
	if ( glInput->isKeyPressed( GKEY_RIGHT ) && timerRight <= glTime->getTime() )
	{
		// do nothing, currently

		timerRight = glTime->getTime() + GAME_TIMER_RIGHTLEFT;
	}

	if ( glInput->isKeyPressed( GKEY_LEFT ) && timerLeft <= glTime->getTime() )
	{
		// do nothing, currently

		timerLeft = glTime->getTime() + GAME_TIMER_RIGHTLEFT;
	}

	if ( glInput->isKeyPressed( GKEY_UP ) && timerUp <= glTime->getTime() )
	{
		// decrease the letter value (GKEY_UP and _DOWN may be switched here...)
		name[nameState]--;
		if ( name[nameState] < 'A' )	// if we're below A, go back to Z
			name[nameState] = 'Z';

		// remind us now, it just changed
		timerLetterRemind = 0.0f;

		// we can't press this button again for a little while (otherwise would happen 60+ times a second)
		timerUp = glTime->getTime() + GAME_TIMER_UPDOWN;
	}

	if ( glInput->isKeyPressed( GKEY_DOWN ) && timerDown <= glTime->getTime() )
	{
		// increase the letter value
		name[nameState]++;
		if ( name[nameState] > 'Z' )	// if we're above Z, go back to A
			name[nameState] = 'A';	
		
		// remind us now, it just changed
		timerLetterRemind = 0.0f;

		// we can't press this button again for a little while (otherwise would happen 60+ times a second)
		timerDown = glTime->getTime() + GAME_TIMER_UPDOWN;
	}

	if ( timerLetterRemind < glTime->getTime() )
	{
		ITextToSpeech speaker;
		string s = "";

		// create a string of the form "A.A.A."
		char c[6];
		c[0] = name[0];
		c[1] = '.';
		c[2] = name[1];
		c[3] = '.';
		c[4] = name[2];
		c[5] = '.';
		s.append(c,6);
		
		// speak the string
		speaker.speakText(s);

		// we'll remind again in a little while
		timerLetterRemind = glTime->getTime() + GAME_TIMER_LETTER;
	}
}

//
// LevelSuccessThink
//  ... called by: Think
//	... description: think code while processing high scores
//
void CGameControl :: LevelSuccessThink()
{
	if ( !bLevelEnd )
	{
		IPlayStationarySound(NULL, 30, SOUND_EVENT_LEVELEND );
		bLevelEnd = true; // don't call these sounds again
		timeTillSayScores = glTime->getTime() + GAME_TIMER_SAYSCORE;	// wait until we've said what the player's score is
		timeTillCanContinue = glTime->getTime() + GAME_TIMER_SAYSCORE + 1.0f; // just after the score has been said
	}

	if ( timeTillSayScores < glTime->getTime() )
	{
		string s;

		// Say the integer value of the player's score
		TCHAR szMsg[MAX_PATH] = TEXT("");
		sprintf( szMsg, TEXT("%i" ), score );
		s.append(szMsg,sizeof(szMsg));
		ITextToSpeech speaker;
		speaker.speakText(s);
		
		timeTillSayScores = glTime->getTime() + 10.0f;

		IPlayStationarySound(NULL, 30, SOUND_EVENT_STARTNEXT );
	}

	if ( (glInput->isKeyPressed( GKEY_SPACE ) || glInput->isKeyPressed( GKEY_ESC )) && timeTillCanContinue < glTime->getTime() )
	{
		gameState = GAME_LEVELINIT;
		CleanupEnts(); // release the ent array, we'll build a new one
		level++;	// next level
		bLevelEnd = false; // reset this switch for the next time around
	}
}


//
// DeathThink
//  ... called by: Think
//	... description: think code while dead
//
void CGameControl :: DeathThink()
{
	if ( (glInput->isKeyPressed( GKEY_ESC ) || glInput->isKeyPressed( GKEY_SPACE )) && timerDeathTransition < glTime->getTime() )
	{
		// go to quit mode, and cleanup memory
		gameState = GAME_QUIT;
		IFreeRacingSamples(); // free the sounds
		CleanupEnts(); // erase the CBaseEntitys
		CleanupLoader(); // make loader clean its database of segments up
		if ( m_pHighScores )
			delete m_pHighScores;
	}

	// PLAY SOUND INSTRUCTION
	if ( timerDeathPlayCredits < glTime->getTime() )
	{
		//IPlayStationarySound(NULL,100,SOUND_EVENT_RETURNTOMENU);
		IPlayStationarySound(NULL,100,SOUND_STREAM_CREDITS);

		timerDeathPlayCredits = glTime->getTime() + GAME_TIMER_DEATHREMIND;
	}
}

//
// DoPhysics
//
//
void CGameControl :: DoPhysics()
{
	if ( gameState == GAME_RACING )
	{
		CBaseEntity* pEnt;
		for ( int i = 0; i < entCount; i++ )
		{
			pEnt = entArray[i];
			pEnt->Physics(); // physics call to every entity, every frame
		}
	}
}

//
// Draw
//	... called by: CUberControl::Draw()
//	... description: handles drawing when we're in game mode
//
void CGameControl :: Draw()
{
	// NOTE: THIS IS VESTIGIAL FROM THE GL DAYS
	/*
	if ( gameState == GAME_INIT )
	{
		// special case, loading splash screen
		//IDrawLoading();
		return;
	}
	*/
}

//
//	PlaySounds
//
//
//
void CGameControl :: PlaySounds()
{
	if ( gameState == GAME_STARTING || gameState == GAME_RACING )
	{
		// we  need to find the player
		for ( int i = 0; i < entCount; i++ )
		{
			if ( entArray[i]->isPlayer() )
			{
				IUpdateSounds( entArray, entCount, entArray[i]->pvars ); // update ALL sounds, using players' position/velocity
				return;
			}
		}
	}
}

//
// CheckCollisions
//	... called by: CGameControl::Think()
//	... description: check for collisions
//
void CGameControl :: CheckCollisions( CBaseEntity *pEnt )
{
	
	for ( int i = 0; i < entCount; i++ )
	{
		CBaseEntity* pSecEnt = entArray[i];

		// are we physical?
		if ( pSecEnt->pvars->ent_flags & ENT_COLLIDABLE )
		{
			// currently not even using hitboxes--this is just a radius check
			Vec3d dist = pEnt->pvars->origin - pSecEnt->pvars->origin;
			if ( dist.Length() < (pEnt->pvars->radius + pSecEnt->pvars->radius ) )
			{
				// we have hit someone
				pEnt->Hit( pSecEnt );
				pSecEnt->Hit( pEnt );
			}
		}
	}
}

//
// LevelLoadData
//	... called by: CGameControl::Think()
//	... description: link to level loading system. fill loader with data from the file
//
void CGameControl :: LevelLoadData()
{
	
	loader->Load(); // have ILevelLoader load the file into its database
	
	gameState = GAME_LEVELINIT;
}

//
// LevelGenerate()
//	... called by: CGameControl::Think()
//	... description: link to level loading system. fill array with entities
//
void CGameControl :: LevelGenerate()
{
	entArray = loader->Generate(level); // generate an array of CBaseEntities
	entCount = loader->getEntCount();	// get the new entcount
	levelLength = loader->getZEnd();	// get the end of the level to check for level success

	// update the CPlayer object with our current amount of lives
	for ( int i = 0; i < entCount; i++ )
	{
		if ( entArray[i]->isPlayer() )
		{
			CPlayer *pPlayer = (CPlayer*)entArray[i];
			pPlayer->lives = lives;
			break;
		}
	}

	gameState = GAME_STARTING; // start the game
	timeTillStart = glTime->getTime() + TIME_TILL_START; // how long is the interval before the race starts?
}



//
// CalcScore()
//	... called by: CGameControl::RaceThink()
//	... description: calculate score/bonustime
//
void CGameControl :: CalcScore()
{
	// first, multiply the coins gotten in this level by the number of points per each
	int addScore = 0;
	addScore = coins*SCORE_PER_COIN;
	
	// get the difference between low and high times
	float possibleBonus = (levelLength/MIN_PLAYER_SPEED - ACCEL_TIME/2) - (levelLength/MAX_PLAYER_SPEED - ACCEL_TIME/2);
	// get the difference between low and our time -- this does not take into account how much you were hit, so it is possible to get no bonus
	bonusTime = (levelLength/MIN_PLAYER_SPEED - ACCEL_TIME/2) - levelDoneTime - pauseTime;

	if ( bonusTime > 0 && lives > 0 )
	{
		// equation: the ratio of our bonus time to the possible is multiplied by the range, and added by 1 to make sure we're only increasing score
		float bonusMultiplier = 1.0f + SCORE_MULTIPLIER_RANGE * bonusTime/possibleBonus;
		addScore = (int)addScore*bonusMultiplier;
		score += addScore;
	}
	else
		score += addScore; // no bonus, just give them the coin points
}

//
// SpeakLevelInfo
//	... called by: Think
//	...	description: say what level we're starting
//
void CGameControl :: SpeakLevelInfo( )
{
	string s;

	// Says "Level X" where x is CGameControl::level
	TCHAR szMsg[MAX_PATH] = TEXT("");
	sprintf( szMsg, TEXT("Level %i" ), level );
	s.append(szMsg,sizeof(szMsg));
	ITextToSpeech speaker;
	speaker.speakText(s);
}

//
// WarningThink
//	... called by: CCoin::WarningThink()
//	...	description: when we approach an obstacle, play warning beeps with increasing frequency
//
// *** NOTE: THIS DID NOT GO INTO THE RELEASE
//
void CGameControl :: WarningThink()
{
	CBaseEntity* pEnt = NULL;
	CPlayer* pPlayer = NULL;
	
	// get the player
	for ( int i = 0; i < entCount; i++ )
	{
		pEnt = entArray[i];

		if( pEnt->isPlayer() )
		{
			pPlayer = (CPlayer*)pEnt;
			break;
		}
	}
	
	CBaseEntity* pClosest = NULL;
	for ( int i = 0; i < entCount; i++ )
	{
		pEnt = entArray[i];
		
		// if it's a mine...
		if ( pEnt->pvars->obj_type == OBJ_OBSTACLE && !(pEnt->pvars->ent_flags & ENT_DEAD ) )
		{
			// if it's in our lane..
			if ( pEnt->pvars->origin.x == pPlayer->pvars->origin.x )
			{
				// if it's ahead of us...
				if ( pEnt->pvars->origin.z > pPlayer->pvars->origin.z )
				{
					// if we don't have anything yet
					if ( !pClosest )
					{
						pClosest = pEnt;
					}
					else 
					{
						// if this is closer than anything else yet
						if ( (pEnt->pvars->origin - pPlayer->pvars->origin).Length() < (pClosest->pvars->origin - pPlayer->pvars->origin).Length() )
						{
							pClosest = pEnt;
						}
					}
				}
			}
		}
	}

	// pClosest should now be the closest obstacle in front of us.
	if ( !pClosest )
		return; // nothing is in front of us, apparently

	Vec3d pointer = pClosest->pvars->origin - pPlayer->pvars->origin;
	float dist = pointer.Length();
	
	// t = d/v
	float time = dist/pPlayer->pvars->speed;

	float klaxInterval = -1;
	float volume = 50;
	
	if ( time >= 5 )
	{
		klaxInterval = -1;
	}
	if ( time < 5 && time >= 4 )
	{
		klaxInterval = 1.0;
	}
	if ( time < 4 && time >= 3 )
	{
		klaxInterval = 0.8;
	}
	if ( time < 3 && time >= 2 )
	{
		klaxInterval = 0.6;
	}
	if ( time < 2 && time >= 1 )
	{
		klaxInterval = 0.4;
	}
	if ( time < 1 )
	{
		klaxInterval = 0.2;
	}

	// determine what the warning level/klaxon interval is
	if ( klaxInterval < 0 )
		return;

	// time to play the sound!
	if ( glTime->getTime() - klaxonLast >= klaxInterval )
	{
		IPlayStationarySound( NULL, volume, SOUND_EVENT_WARNING );
		klaxonLast = glTime->getTime();
	}

}