
#include "CPlayer.h"
#include "Vec3.h"
#include "globalconst.h"
#include "stateflagdefs.h"
#include "CBaseEntity.h"
#include "ITime.h"
#include "IInput.h"
#include "soundfunc.h"

//-----------------------------------------------------------
// Class CPlayer
//
//	This class represents the player, it is the most complicated
// object in the game for that reason. 
//
//
//	UPDATES:
//	2/11/04 - Daniel: created, spawn, think (except for that 
//				which requires input), hit
//	2/18/04 - Daniel: added input to think
//	2/27/04 - Daniel: made the switching smooth
//	3/15/04 - Daniel: put player input in its own function
//	3/21/04 - Daniel: now checks the switching every frame, so it's smooth.
//  4/10/04 - Daniel: hit happens immediately, doesn't wait for think. Think is becoming less useful...
//	4/13/04 - Daniel: sound is back in switching, can accelerate while switching, and a delay between switching so not to go too far
//
//-----------------------------------------------------------

extern IInput *glInput;

//
// Constructor, default
//
//
CPlayer :: CPlayer()
{
}


//
// Spawn
//	... called by: ILevelLoader::Generate
//	... description: initialize pvars to what this object requires
//
void CPlayer :: Spawn()
{
	// object-specific traits in entvars
	pvars->obj_type				= OBJ_PLAYER; // object type
	pvars->player_state			= PL_ACCELERATING; // we're at the beginning state
	pvars->ent_flags			= 0;			 // initialize
	pvars->ent_flags			= ((((( pvars->ent_flags | ENT_COLLIDABLE )
									| ENT_VISIBLE )
									| ENT_AUDIBLE )
									| ENT_MOVABLE )
									| ENT_PLAYER  ); // flags
	pvars->radius				= RADIUS_PLAYER; // radius for collision
	pvars->speed				= 0;
	pvars->thinktime			= glTime->getTime();
	pvars->velocity.z			= 1; // head downfield
	pvars->velocity.x			= 0; 
	pvars->velocity.y			= 0;
	pvars->orientation = pvars->velocity;
	pvars->switch_state			= SWITCH_NO;
	// origin handled once, when created, and outside this function

	// player variables
	lives		= NUM_LIVES_STANDARD; // give them some lives
	highSpeed	= 0; // start not moving
	lowSpeed	= 0; // start not moving
	coins = 0;
	stunnedTime = 0;
	level = 0;
	timerLastSwitchEnd = 0;
}

//
//	Think
//	... called by: CGameControl->Think()
//	...	description: do game logic at intervals, check for updates/state change, etc.
//
void CPlayer :: Think()
{
	// do this first, before we jump the gun on other stuff
	if ( pvars->player_state == PL_HIT )
	{
		// code moved to Hit, now happens immediately
		return;
	}

	// stunned -- we COULD have done this by just making thinktime 5 seconds or so ahead, but there's the
	// possibility something might want to be done during this time, and there're initializations to be made
	// we don't necessarily want in the more-general-purpose PL_ACCELERATING state.
	if ( pvars->player_state == PL_STUNNED )
	{
		if ( stunnedTime < glTime->getTime() ) // time to get back in the game
		{
			pvars->ent_flags = pvars->ent_flags | ENT_MOVING;
			pvars->player_state = PL_ACCELERATING;

			// vector stuff
			pvars->ang_vel.x = 0; // no more spinning
			pvars->orientation.z = 1; // head downfield
			pvars->orientation.x = 0; 
			pvars->orientation.y = 0; 
		}
	}
	
	// player is getting (back) up to full speed
	if ( pvars->player_state == PL_ACCELERATING ) 
	{
		if ( lowSpeed < MIN_PLAYER_SPEED )
		{
			// equation: acceleration times fraction of a second this think represents
			// ... could also be done with a target time compared to glTime, but this works fine
			lowSpeed += (MIN_PLAYER_SPEED/ACCEL_TIME) * NT_PLAYER;
			if ( pvars->speed < lowSpeed )
				pvars->speed = lowSpeed; // they must go at least the minimum speed, since it's changing
		}
		if ( highSpeed < MAX_PLAYER_SPEED )
		{
			// equation: acceleration times fraction of a second this think represents
			// ... could also be done with a target time compared to glTime, but this works fine
			highSpeed += (MAX_PLAYER_SPEED/ACCEL_TIME) * NT_PLAYER;
		}
		if ( lowSpeed >= MIN_PLAYER_SPEED && highSpeed >= MAX_PLAYER_SPEED )
		{
			// no more reason to accelerate, we're at our required speeds
			pvars->player_state = PL_STRAIGHTAWAY;
		}

		pvars->ent_flags = pvars->ent_flags | ENT_MOVING; // we're moving, need to do collisions
			
	}
	// when is the next think time?
	pvars->thinktime = glTime->getTime() + NT_PLAYER;
}

//
// SwitchBegin()
//	... called by: HandleInput()
//	... description: begins the switch physics
//
void CPlayer :: SwitchBegin()
{
	// play a sound off to our right or left
	Vec3d loc = pvars->origin + pvars->velocity * pvars->speed * glTime->reportFrameInterval();
	float pos[3] = { switchDestinationLane, pvars->origin.y, loc.z };
	float vel[3] = { pvars->velocity.x, pvars->velocity.y, pvars->velocity.z };
	IPlay3DSound( pos, vel, 200, SOUND_3D_MOVING );

	//Eden: Now says what lane you're in
	if ( level <= 5 ) // .. but only on and below level 5
	{
		if( switchDestinationLane == TRACK_3_X )
			IPlayStationarySound( NULL, 60, SOUND_EVENT_LEFTLANE );
		else if( switchDestinationLane == TRACK_2_X )
			IPlayStationarySound( NULL, 60, SOUND_EVENT_CENTERLANE );
		else if( switchDestinationLane == TRACK_1_X )
			IPlayStationarySound( NULL, 60, SOUND_EVENT_RIGHTLANE );
	}
}

//
// SwitchCheck()
//	... called by: 
//	... description: 
//
void CPlayer :: SwitchCheck()
{
	if ( pvars->switch_state != SWITCH_NO )
	{
		// move us in the direction we need to go
		pvars->origin.x += ((switchDestinationLane - switchInitialLane )/SWITCH_INTERVAL)*glTime->reportFrameInterval();

		// check to see if we've gone too far
		if ( pvars->switch_state == SWITCH_LEFT )
		{	
			if ( pvars->origin.x >= switchDestinationLane )
			{
				// we've gotten to where we're going
				pvars->origin.x = switchDestinationLane; // set our x value to the destination lane
				pvars->switch_state = SWITCH_NO; // we're no longer switching
				IPlayStationarySound( NULL, 5, SOUND_EVENT_DONEMOVING ); // play the done moving beep
				timerLastSwitchEnd = glTime->getTime();
			}
		}
		if ( pvars->switch_state == SWITCH_RIGHT )
		{
			if ( pvars->origin.x <= switchDestinationLane )
			{
				// we've gotten to where we're going
				pvars->origin.x = switchDestinationLane;
				pvars->switch_state = SWITCH_NO;
				IPlayStationarySound( NULL, 5, SOUND_EVENT_DONEMOVING );
				timerLastSwitchEnd = glTime->getTime();
			}
		}
	}

}


//
// HandleInput()
//	... called by: CGameControl::RaceThink
//	... description: handles input every frame
//
void CPlayer :: HandleInput()
{
	if ( pvars->player_state == PL_ACCELERATING || pvars->player_state == PL_STRAIGHTAWAY )
	{


		if ( glInput->isKeyPressed( GKEY_UP ) )
		{
			// speed up
			pvars->speed += (MAX_PLAYER_SPEED-MIN_PLAYER_SPEED)/SPEED_INTERVALS * glTime->reportFrameInterval(); // go up by a certain amount
			if ( pvars->speed > highSpeed )
				pvars->speed = highSpeed; // stay at the maximum speed
		}
		if ( glInput->isKeyPressed( GKEY_DOWN ) )
		{
			// slow down
			pvars->speed -= (MAX_PLAYER_SPEED-MIN_PLAYER_SPEED)/SPEED_INTERVALS * glTime->reportFrameInterval(); // go down by a certain amount
			if ( pvars->speed < lowSpeed )
				pvars->speed = lowSpeed; // we have a minimum speed
		}
	
		if ( (glTime->getTime() - timerLastSwitchEnd) >= 0.1f )
		{
			// if we're not pressing anything, we can press something whenever we want to
			if ( !glInput->isKeyPressed( GKEY_LEFT ) && !glInput->isKeyPressed( GKEY_RIGHT ) )
				timerLastSwitchEnd = 0;
		}

		if ( pvars->switch_state == SWITCH_NO && (glTime->getTime() - timerLastSwitchEnd) >= 0.1f )
		{
			if ( glInput->isKeyPressed( GKEY_LEFT ) )
			{
				if ( pvars->origin.x == TRACK_2_X )
				{
					// switch from 2 to 1
					switchInitialLane = TRACK_2_X;
					switchDestinationLane = TRACK_1_X;
					pvars->switch_state = SWITCH_LEFT;
					SwitchBegin();
				}
				if ( pvars->origin.x == TRACK_3_X )
				{
					// switch from 3 to 2
					switchInitialLane = TRACK_3_X;
					switchDestinationLane = TRACK_2_X;
					pvars->switch_state = SWITCH_LEFT;
					SwitchBegin();
				}
			}
			if ( glInput->isKeyPressed( GKEY_RIGHT ) )
			{
				if ( pvars->origin.x == TRACK_2_X )
				{
					// switch from 2 to 3
					switchInitialLane = TRACK_2_X;
					switchDestinationLane = TRACK_3_X;
					pvars->switch_state = SWITCH_RIGHT;
					SwitchBegin();
				}
				if ( pvars->origin.x == TRACK_1_X )
				{
					// switch from 1 to 2
					switchInitialLane = TRACK_1_X;
					switchDestinationLane = TRACK_2_X;
					pvars->switch_state = SWITCH_RIGHT;
					SwitchBegin();
				}
			}
		}
	}
}

//
// Remove()
//	... called by: any
//	... description: removes the player
//
void CPlayer :: Remove()
{
	// nothing here yet
}

//
// Hit
//	... called by: CGameControl->CheckCollisions()
//	... description: do game logic when a object collides with this one
//
void CPlayer :: Hit( CBaseEntity * victim )
{
	// what happens if we hit a coin
	if ( victim->pvars->obj_type == OBJ_COIN ) 
	{
		coins++;
	}
	
	// what happens if we hit an obstacle
	if ( victim->pvars->obj_type == OBJ_OBSTACLE )
	{
		pvars->player_state = PL_HIT;
		if ( lives == 0 ) // we just died for good
		{
			pvars->player_state = PL_DEAD; // CGameControl will pick up on this, and it's curtains
			return;
		}
		lives--; // since this is after the check, "lives" is technically "extra" lives
		pvars->speed = lowSpeed = highSpeed = 0;  // stop moving completely
		pvars->player_state = PL_STUNNED; // we're now stunned for a certain penalty time
		stunnedTime = glTime->getTime() + STUN_TIME_STANDARD; // we're out for this amount of time
		pvars->ent_flags = pvars->ent_flags & ~ENT_MOVING; // we're no longer moving
		pvars->ang_vel.x = (float)0.66*PI;	// spin around, adding insult to injury
		switchOldSpeed = 0;
	}
}


