
#include "glaux.h"
#include "glfunc.h"
#include "globalconst.h"
#include <gl/gl.h>
#include <gl/glu.h>
#include "Vec3.h"

//-----------------------------------------------------------
// Class GTrackArray
//
//	NOTE: This class is not a primary OR secondary requirement, and
// comes courtesy of Daniel's free time.
//
//	With that said, it performs three roles: first, to create a "track"
// out of thin air when given only the length to the last object in the game.
// Secondly, it can warp that track in a few ways, changing the x and y coordinates
// of all track vertices according to some function of z. Finally, it can provide
// similar offsets for the locations of the objects and the player in the game, to be
// used when rendering. See the rendering code for how this works.
//
//-----------------------------------------------------------

//
// Cleanup
//	... called by: any
//	... description: free memory
//
void GTrackArray :: Cleanup()
{
	delete [] track; // delete entire array
	track = NULL; // delete the pointer to the array
}

//
// Create
//	... called by: any
//	... description: create basic track (ground only) of a given length
//
void GTrackArray :: Create( int levelLength )
{
	float z = 0; // used to keep track of where we are
	float y = -4; // this value determines how low the track is below the player
	float x = 0; // used to keep track of which lane we're generating

	float increment = 0; // count up until the end of the level
	int segCount = 0; // how many will that be?

	while ( increment <= levelLength )
	{
		segCount++; // another segment needed
		increment += TRACK_SEG_LENGTH; // add by the seg length
	}

	int lineTotal = 2*segCount; // how many lines inbetween the three main lanes?
	int segTotal = 3*segCount; // how many of the three main lanes?
	int totalTotal = lineTotal+segTotal; // how many quads total?
	track = new GTrackSegment[totalTotal]; // create the array

	// go through all the main lanes
	for ( int i = 0; i < segTotal; i++ )
	{
		switch (i%3)
		{
		case 0:
			x = TRACK_1_X; // do the first lane
			break;
		case 1:
			x = TRACK_2_X; // do the second lane
			break;
		case 2:
			x = TRACK_3_X; // do the third lane
			break;
		}

		// first vertex
		track[i].first[0] = x - 4;
		track[i].first[1] = y;
		track[i].first[2] = z;
		track[i].firstNorm[0] = 0;
		track[i].firstNorm[1] = 1;
		track[i].firstNorm[2] = 0;

		// second vertex
		track[i].second[0] = x - 4;
		track[i].second[1] = y;
		track[i].second[2] = z + TRACK_SEG_LENGTH;
		track[i].secondNorm[0] = 0;
		track[i].secondNorm[1] = 1;
		track[i].secondNorm[2] = 0;

		// third vertex
		track[i].third[0] = x + 4;
		track[i].third[1] = y;
		track[i].third[2] = z + TRACK_SEG_LENGTH;
		track[i].thirdNorm[0] = 0;
		track[i].thirdNorm[1] = 1;
		track[i].thirdNorm[2] = 0;
	
		// fourth vertex
		track[i].fourth[0] = x + 4;
		track[i].fourth[1] = y;
		track[i].fourth[2] = z;
		track[i].fourthNorm[0] = 0;
		track[i].fourthNorm[1] = 1;
		track[i].fourthNorm[2] = 0;

		// not currently used by render,
		// but would create a dark road if textures weren't being used
		track[i].R = 0.1f;
		track[i].G = 0.1f;
		track[i].B = 0.1f;

		track[i].textureIndex = TEXTURE_TRACKMAIN; // use the ground texture

		if ( i%3 == 2 )
			z += TRACK_SEG_LENGTH;
	}

	z = 0; // reset our z value

	int numDone = 0; // need this to make "yellow lines" sync up
	// go through the lines inbetween the main lanes
	for ( int i = segTotal; i < totalTotal; i++ )
	{
		switch ( numDone % 2 )
		{
		case 0:
			x = (TRACK_2_X - TRACK_1_X)/2 + TRACK_1_X; // between first and second
			break;
		case 1:
			x = (TRACK_3_X - TRACK_2_X)/2 + TRACK_2_X; // between second and third
			break;
		}
		
		// first vertex
		track[i].first[0] = x - 1;
		track[i].first[1] = y;
		track[i].first[2] = z;
		track[i].firstNorm[0] = 0;
		track[i].firstNorm[1] = 1;
		track[i].firstNorm[2] = 0;

		// second vertex
		track[i].second[0] = x - 1;
		track[i].second[1] = y;
		track[i].second[2] = z + TRACK_SEG_LENGTH;
		track[i].secondNorm[0] = 0;
		track[i].secondNorm[1] = 1;
		track[i].secondNorm[2] = 0;

		// third vertex
		track[i].third[0] = x + 1;
		track[i].third[1] = y;
		track[i].third[2] = z + TRACK_SEG_LENGTH;
		track[i].thirdNorm[0] = 0;
		track[i].thirdNorm[1] = 1;
		track[i].thirdNorm[2] = 0;

		// fourth vertex
		track[i].fourth[0] = x + 1;
		track[i].fourth[1] = y;
		track[i].fourth[2] = z;
		track[i].fourthNorm[0] = 0;
		track[i].fourthNorm[1] = 1;
		track[i].fourthNorm[2] = 0;

		// should this be a "yellow line"?
		if ( numDone%4 == 2 || numDone%4 == 3 )
		{
			track[i].R = 0.1f;
			track[i].G = 0.1f;
			track[i].B = 0.1f;
			track[i].textureIndex = TEXTURE_LINE2;
		}
		else
		{
			track[i].R = 0.5f;
			track[i].G = 0.5f;
			track[i].B = 0.0f;
			track[i].textureIndex = TEXTURE_LINE1;
		}

		if ( numDone%2 == 1 )
			z += TRACK_SEG_LENGTH; // if we've done two, go up a segment lenth
		
		numDone++; // keep track of how many
	}

	segNum = totalTotal; // update the class' member data
	bCreated = true; // mark the track as created

}

//
// Create2
//	... called by: any
//	... description: create tunnel track of a given length
//
void GTrackArray :: Create2( int levelLength )
{
	float z = 0; // keep track of our z values
	float y = -4; // how much lower than the player is the track?
	float x = 0; // keep track of which lane we're doing

	// see how many segments total we need to make
	float increment = 0;
	int segCount = 0;
	while ( increment <= levelLength )
	{
		segCount++;
		increment += TRACK_SEG_LENGTH;
	}
	
	int facesPerSegment = 10; // for the tunnel, we have the ceiling, the four walls, the three lanes, and the two "yellow lines"
	int totalFaces = segCount * facesPerSegment; // how many quads total?
	int lineTotal = 2*segCount; // two "yellow lines"
	int segTotal = 3*segCount; // three main lanes
	int sideTotal = 4*segCount; // four sides
	int topTotal = 1*segCount; // one ceiling
	int totalTotal = totalFaces; // how many total?
	track = new GTrackSegment[totalTotal]; // create the array
	
	// FIRST, THE MAIN THREE LANE FACES
	for ( int i = 0; i < segTotal; i++ )
	{
		switch (i%3)
		{
		case 0:
			x = TRACK_1_X; // do the first lane
			break;
		case 1:
			x = TRACK_2_X; // do the second lane
			break;
		case 2:
			x = TRACK_3_X; // do the third lane
			break;
		}
		
		// first vertex
		track[i].first[0] = x - 4;
		track[i].first[1] = y;
		track[i].first[2] = z;
		track[i].firstNorm[0] = 0;
		track[i].firstNorm[1] = 1;
		track[i].firstNorm[2] = 0;

		// second vertex
		track[i].second[0] = x - 4;
		track[i].second[1] = y;
		track[i].second[2] = z + TRACK_SEG_LENGTH;
		track[i].secondNorm[0] = 0;
		track[i].secondNorm[1] = 1;
		track[i].secondNorm[2] = 0;

		// third vertex
		track[i].third[0] = x + 4;
		track[i].third[1] = y;
		track[i].third[2] = z + TRACK_SEG_LENGTH;
		track[i].thirdNorm[0] = 0;
		track[i].thirdNorm[1] = 1;
		track[i].thirdNorm[2] = 0;

		// fourth vertex
		track[i].fourth[0] = x + 4;
		track[i].fourth[1] = y;
		track[i].fourth[2] = z;
		track[i].fourthNorm[0] = 0;
		track[i].fourthNorm[1] = 1;
		track[i].fourthNorm[2] = 0;
	
		// NOT USED CURRENTLY
		track[i].R = 1;
		track[i].G = 1;
		track[i].B = 1;

		track[i].textureIndex = TEXTURE_TRACKMAIN; // use the ground texture

		if ( i%3 == 2 )
			z += TRACK_SEG_LENGTH; // if we've done three, go up one
	}

	z = 0;

	// THE LINE FACES (INBETWEEN LANES)
	int numDone = 0; // can't just use i, since it might be off by one
	for ( int i = segTotal; i < segTotal+lineTotal; i++ )
	{
		switch ( numDone % 2 )
		{
		case 0:
			x = (TRACK_2_X - TRACK_1_X)/2 + TRACK_1_X; // between first and second
			break;
		case 1:
			x = (TRACK_3_X - TRACK_2_X)/2 + TRACK_2_X; // between second and third
			break;
		}
	
		// first vertex
		track[i].first[0] = x - 1;
		track[i].first[1] = y;
		track[i].first[2] = z;
		track[i].firstNorm[0] = 0;
		track[i].firstNorm[1] = 1;
		track[i].firstNorm[2] = 0;

		// second vertex
		track[i].second[0] = x - 1;
		track[i].second[1] = y;
		track[i].second[2] = z + TRACK_SEG_LENGTH;
		track[i].secondNorm[0] = 0;
		track[i].secondNorm[1] = 1;
		track[i].secondNorm[2] = 0;

		// third vertex
		track[i].third[0] = x + 1;
		track[i].third[1] = y;
		track[i].third[2] = z + TRACK_SEG_LENGTH;
		track[i].thirdNorm[0] = 0;
		track[i].thirdNorm[1] = 1;
		track[i].thirdNorm[2] = 0;

		// fourth vertex
		track[i].fourth[0] = x + 1;
		track[i].fourth[1] = y;
		track[i].fourth[2] = z;
		track[i].fourthNorm[0] = 0;
		track[i].fourthNorm[1] = 1;
		track[i].fourthNorm[2] = 0;

		if ( numDone%4 == 2 || numDone%4 == 3 )
		{
			// not currenlty used
			track[i].R = 1;
			track[i].G = 1;
			track[i].B = 1;
			track[i].textureIndex = TEXTURE_LINE2; // second type of line
		}
		else
		{
			track[i].R = 1.0f;
			track[i].G = 1.0f;
			track[i].B = 0.0f;
			track[i].textureIndex = TEXTURE_LINE1; // first type of line
		}
		
		
		if ( numDone%2 == 1 )
			z += TRACK_SEG_LENGTH; // if we've done two, go to the next z value

		numDone++; // keep track of this
	}

	z = 0; // reset z value
	numDone = 0; // reset counter
	// NEXT, THE TOP FACE
	for ( int i = segTotal+lineTotal; i < segTotal+lineTotal+topTotal; i++ )
	{
		x = 0; // always in the center
		y = 16; // always this much above us

		// first vertex
		track[i].first[0] = x - 14;
		track[i].first[1] = y;
		track[i].first[2] = z;
		track[i].firstNorm[0] = 0;
		track[i].firstNorm[1] = -1;
		track[i].firstNorm[2] = 0;

		// second vertex
		track[i].second[0] = x + 14;
		track[i].second[1] = y;
		track[i].second[2] = z ;
		track[i].secondNorm[0] = 0;
		track[i].secondNorm[1] = -1;
		track[i].secondNorm[2] = 0;

		// third vertex
		track[i].third[0] = x + 14;
		track[i].third[1] = y;
		track[i].third[2] = z + TRACK_SEG_LENGTH;
		track[i].thirdNorm[0] = 0;
		track[i].thirdNorm[1] = -1;
		track[i].thirdNorm[2] = 0;

		// fourth vertex
		track[i].fourth[0] = x - 14;
		track[i].fourth[1] = y;
		track[i].fourth[2] = z + TRACK_SEG_LENGTH;
		track[i].fourthNorm[0] = 0;
		track[i].fourthNorm[1] = -1;
		track[i].fourthNorm[2] = 0;
		
		track[i].textureIndex = TEXTURE_CEILING; // use the ceiling texture

		// ONLY FOR NON-TEXTURE VERSIONS
		if ( numDone%2 == 1  )
		{
			track[i].R = 0.1f;
			track[i].G = 0.1f;
			track[i].B = 0.2f;
		}
		else
		{
			track[i].R = 0.1f;
			track[i].G = 0.1f;
			track[i].B = 0.6f;
		}

		z += TRACK_SEG_LENGTH; // increase our z value every time
	}

	x = 0;
	z = 0;
	float topy = 16; // the height of the tunnel above 0
	float boty = -4; // the floor
	float midy = 6; // the middle
	float firstx; // the first x value of a quad
	float secx; // second x value of a quad
	float lowy; // lower y value of a quad
	float highy; // high y value of a quad
	float firstz; // first (lower) z value of a quad
	float secz; // second (higher) z value of a quad
	float sin45deg = 0.7071f;
	float normalx = 0; // for our normal vector
	float normaly = 0; // for our normal vector
	// NEXT, the SIDES
	numDone = 0;
	for ( int i = segTotal+lineTotal+topTotal; i < totalTotal; i++ )
	{
		switch (numDone%4)
		{
		case 0: // LEFT,LOWER
			firstx = x - 14;
			secx = x - 18;
			lowy = boty;
			highy = midy;
			normalx = sin45deg;
			normaly = sin45deg;
			firstz = z;
			secz = z+TRACK_SEG_LENGTH;
			track[i].textureIndex = TEXTURE_SIDELOW;
			break;
		case 1: // LEFT, UPPER
			firstx = x - 18;
			secx = x - 14;
			lowy = midy;
			highy = topy;
			normalx = sin45deg;
			normaly = -sin45deg;
			firstz = z;
			secz = z+TRACK_SEG_LENGTH;
			track[i].textureIndex = TEXTURE_SIDEHIGH;
			break;
		case 2: // RIGHT, LOWER
			firstx = x + 14;
			secx = x + 18;
			lowy = boty;
			highy = midy;
			normalx = -sin45deg;
			normaly = sin45deg;
			firstz = z+TRACK_SEG_LENGTH;
			secz = z;
			track[i].textureIndex = TEXTURE_SIDELOW;
			break;
		case 3: // RIGHT, UPPER
			firstx = x + 18;
			secx = x + 14;
			lowy = midy;
			highy = topy;
			normalx = -sin45deg;
			normaly = -sin45deg;
			firstz = z+TRACK_SEG_LENGTH;
			secz = z;
			track[i].textureIndex = TEXTURE_SIDEHIGH;
			break;
		}

		// first vertex generated from the above switch statement
		track[i].first[0] = firstx;
		track[i].first[1] = lowy;
		track[i].first[2] = firstz;
		track[i].firstNorm[0] = normalx;
		track[i].firstNorm[1] = normaly;
		track[i].firstNorm[2] = 0;

		// second vertex generated from the above switch statement
		track[i].second[0] = secx;
		track[i].second[1] = highy;
		track[i].second[2] = firstz;
		track[i].secondNorm[0] = normalx;
		track[i].secondNorm[1] = normaly;
		track[i].secondNorm[2] = 0;

		// third vertex generated from the above switch statement
		track[i].third[0] = secx;
		track[i].third[1] = highy;
		track[i].third[2] = secz;
		track[i].thirdNorm[0] = normalx;
		track[i].thirdNorm[1] = normaly;
		track[i].thirdNorm[2] = 0;

		// fourth vertex generated from the above switch statement
		track[i].fourth[0] = firstx;
		track[i].fourth[1] = lowy;
		track[i].fourth[2] = secz;
		track[i].fourthNorm[0] = normalx;
		track[i].fourthNorm[1] = normaly;
		track[i].fourthNorm[2] = 0;

		// FOR NON-TEXTURE VERSIONS ONLY
		if ( numDone%8 <= 3 )
		{
			if ( numDone%4 == 1 || numDone%4 == 3  )
			{
				track[i].R = 0.5f;
				track[i].G = 0.5f;
				track[i].B = 0.6f;
			}
			else
			{
				track[i].R = 0.5f;
				track[i].G = 0.5f;
				track[i].B = 0.1f;
			}
		}
		else
		{
			if ( numDone%4 == 0 || numDone%4 == 2  )
			{
				track[i].R = 0.5f;
				track[i].G = 0.5f;
				track[i].B = 0.6f;
			}
			else
			{
				track[i].R = 0.5f;
				track[i].G = 0.5f;
				track[i].B = 0.1f;
			}
		}

		if ( numDone%4 == 3 )
			z += TRACK_SEG_LENGTH;

		numDone++;
	}

	segNum = totalTotal; // update member data
	bCreated = true; // flag us as created

}

//
// Draw
//	... called by: NONE
//	... description: openGL only, Direct3D handles track-drawing itself
//
void GTrackArray :: Draw()
{
	
	/*
	for ( int i = 0; i < segNum; i++ )
	{
		glColor3f( track[i].R, track[i].G, track[i].B);
		glBegin(GL_QUADS);
			glVertex3f(track[i].first[0], track[i].first[1], track[i].first[2]);
			glVertex3f(track[i].second[0], track[i].second[1], track[i].second[2]);
			glVertex3f(track[i].third[0], track[i].third[1], track[i].third[2]);
			glVertex3f(track[i].fourth[0], track[i].fourth[1], track[i].fourth[2]);
		glEnd(); 
	}
	*/
}

//
// DoWarp
//	... called by: currently, CMyD3dApplication::FrameMove
//	... description: given a warp type and radius/fraction information, call a warp function
//
void GTrackArray :: DoWarp( float radius, float frac, int type )
{
	// radius is how much we multiply the sin/cos in the warp functions by,
	// frac is what we multiply the distance by to relate it to angle.
	warpType = type; // update the member data
	switch ( warpType )
	{
		case WARP_NONE: // do nothing
		case WARP_TUNNEL: // do nothing -- this is stub of something used to be used to test the track
			break;
		case WARP_PERPSPIRAL: // this creates a spiral where everything still always points up
			Warp(radius,frac,radius,frac);
			break;
		case WARP_AXIALSPIRAL: // this creates a spiral where "up" is towards the axis
			Warp2(radius,frac);
			break;
		case WARP_CURVY: // this creates a roller-coaster-esque up-and-down warp
			Warp3(radius,frac);
			break;
	}
}


//
// Warp
//	... called by: DoWarp
//	... description: PERPSPIRAL warp, goes in a spiral but the normal of the track remains constant
//
void GTrackArray :: Warp( float outSin, float inSin, float outCos, float inCos )
{
	// NOTE: it would be possible to have the sin and cos components altered differently from each other,
	// but this is not currently being used.

	float zAxisAngle = 0; // initialize
	for ( int i = 0; i < segNum; i++ ) // go through each segment
	{
		zAxisAngle = 2*PI*inCos*track[i].first[2]; // this is the perpspiral equation
		track[i].first[0] = track[i].first[0] + outCos * cos(zAxisAngle); // offset
		track[i].first[1] = track[i].first[1] + outSin * sin(zAxisAngle); // offset
		
		zAxisAngle = 2*PI*inCos*track[i].second[2]; // this is the perpspiral equation
		track[i].second[0] = track[i].second[0] + outCos * cos(zAxisAngle); // offset
		track[i].second[1] = track[i].second[1] + outSin * sin(zAxisAngle); // offset
	
		zAxisAngle = 2*PI*inCos*track[i].third[2]; // this is the perpspiral equation
		track[i].third[0] = track[i].third[0] + outCos * cos(zAxisAngle); // offset
		track[i].third[1] = track[i].third[1] + outSin * sin(zAxisAngle); // offset

		zAxisAngle = 2*PI*inCos*track[i].fourth[2]; // this is the perpspiral equation
		track[i].fourth[0] = track[i].fourth[0] + outCos * cos(zAxisAngle); // offset
		track[i].fourth[1] = track[i].fourth[1] + outSin * sin(zAxisAngle); // offset
	}
}

//
// Warp2
//	... called by: DoWarp
//	... description: AXIALSPIRAL warp, goes in a spiral, and the normal of the track points towards the axis
//		NOTE: I have no doubt this is horribly inefficient.
//
void GTrackArray :: Warp2( float radius, float frac )
{
	float zAxisAngleOff = 0; // how much should we offset the angle?
	float zAxisAngle = 0; // what is the angle currently compared to the AXIS? (what is our vertex in polar coordinates?)
	float dist = 0;
	Vec3d radial(0,radius,0); // a vector from 0,0,0 to the axis of the warp
	Vec3d current; // what's the current vector?
	Vec3d neo; // vector from the axis to the new point
	Vec3d final; // final vector
	Vec3d tem; // temporary vector
	for ( int i = 0; i < segNum; i++ )
	{
		// first vertex
		current.x = track[i].first[0]; // load in the first into current
		current.y = track[i].first[1];
		current.z = track[i].first[2];
		radial.z = track[i].first[2]; // change the radial vector's z value
		neo = current - radial; // radial to our track vector
		tem = radial - current; // and the opposite
		zAxisAngle = atan2( neo.y,neo.x); // the current angle, i.e. our vertex in polar coordinates
		zAxisAngleOff = 2*PI*frac*current.z; // AXIALSPIRAL EQUATION
		zAxisAngle += zAxisAngleOff; // add to the angle
		dist = neo.Length(); // what is the distance from this vertex to the axis?
		neo.x = cos(zAxisAngle); // extract the x value from our angle
		neo.y = sin(zAxisAngle); // extract the y value from our angle
		neo = neo * dist; // multiply by the length to the axis
		final = neo+tem; // NEO NOW REPRESENTS VECTOR FROM AXIS TO NEW POINT, TEM REPRESENTS FROM OLD POINT TO AXIS
		track[i].first[0] += final.x; // add the offset
		track[i].first[1] += final.y; // add the offset

		// second vertex
		current.x = track[i].second[0];
		current.y = track[i].second[1];
		current.z = track[i].second[2];
		radial.z = track[i].second[2];
		neo = current - radial;
		tem = radial - current;
		zAxisAngle = atan2( neo.y,neo.x);
		zAxisAngleOff = 2*PI*frac*current.z;
		zAxisAngle += zAxisAngleOff;
		dist = neo.Length();
		neo.x = cos(zAxisAngle);
		neo.y = sin(zAxisAngle);
		neo = neo * dist;
		final = neo+tem;
		track[i].second[0] += final.x;
		track[i].second[1] += final.y;

		// third vertex
		current.x = track[i].third[0];
		current.y = track[i].third[1];
		current.z = track[i].third[2];
		radial.z = track[i].third[2];
		neo = current - radial;
		tem = radial - current;
		zAxisAngle = atan2( neo.y,neo.x);
		zAxisAngleOff = 2*PI*frac*current.z;
		zAxisAngle += zAxisAngleOff;
		dist = neo.Length();
		neo.x = cos(zAxisAngle);
		neo.y = sin(zAxisAngle);
		neo = neo * dist;
		final = neo + tem;
		track[i].third[0] += final.x;
		track[i].third[1] += final.y;

		// four vertex
		current.x = track[i].fourth[0];
		current.y = track[i].fourth[1];
		current.z = track[i].fourth[2];
		radial.z = track[i].fourth[2];
		neo = current - radial;
		tem = radial - current;
		zAxisAngle = atan2( neo.y,neo.x);
		zAxisAngleOff = 2*PI*frac*current.z;
		zAxisAngle += zAxisAngleOff;
		dist = neo.Length();
		neo.x = cos(zAxisAngle);
		neo.y = sin(zAxisAngle);
		neo = neo * dist;
		final = neo+tem;
		track[i].fourth[0] += final.x;
		track[i].fourth[1] += final.y;

	}
}

//
// Warp3
//	... called by: DoWarp
//	... description: CURVY warp, goes up and down like a rollercoaster
//
void GTrackArray :: Warp3(float outSin, float frac )
{

	// go through each
	for ( int i = 0; i < segNum; i++ )
	{
		track[i].first[1] += getCurvyYOffset(track[i].first[2],outSin,frac); // go to a handy function
		track[i].second[1] += getCurvyYOffset(track[i].second[2],outSin,frac);
		track[i].third[1] += getCurvyYOffset(track[i].third[2],outSin,frac);
		track[i].fourth[1] += getCurvyYOffset(track[i].fourth[2],outSin,frac);
	}
}

//
// getAxialXOffset
//	... called by: GTrackArray::getXOffsetFromXYZ
//	... description: for AXIAL_SPIRAL warp, get the x offset based on x, y, and z of the original point
//
float GTrackArray::getAxialXOffset( float x, float y, float z, float radius, float frac )
{
	float give; // return value

	float zAxisAngleOff = 0;
	float zAxisAngle = 0;
	float dist = 0;
	Vec3d radial(0,radius,0);
	Vec3d current;
	Vec3d neo;
	Vec3d final;
	Vec3d tem;

	// SEE THE AXIAL SPIRAL WARP, THIS IS THE SAME, BUT WE ONLY CARE ABOUT THE FINAL X VALUE
	current.x = x;
	current.y = y;
	current.z = z;
	radial.z = z;
	neo = current - radial;
	tem = radial - current;
	zAxisAngle = atan2( neo.y,neo.x);
	zAxisAngleOff = 2*PI*frac*current.z;
	zAxisAngle += zAxisAngleOff;
	dist = neo.Length();
	neo.x = cos(zAxisAngle);
	neo.y = sin(zAxisAngle);
	neo = neo * dist;
	final = neo+tem;
	give = final.x;

	return give;
}

//
// getAxialYOffset
//	... called by: GTrackArray::getYOffsetFromXYZ
//	... description: for AXIAL_SPIRAL warp, get the y offset based on x, y, and z of the original point
//
float GTrackArray::getAxialYOffset( float x, float y, float z, float radius, float frac )
{
	float give; // return value

	float zAxisAngleOff = 0;
	float zAxisAngle = 0;
	float dist = 0;
	Vec3d radial(0,radius,0);
	Vec3d current;
	Vec3d neo;
	Vec3d final;
	Vec3d tem;

	// SEE THE AXIAL SPIRAL WARP, THIS IS THE SAME, BUT WE ONLY CARE ABOUT THE FINAL Y VALUE
	current.x = x;
	current.y = y;
	current.z = z;
	radial.z = z;
	neo = current - radial;
	tem = radial - current;
	zAxisAngle = atan2( neo.y,neo.x);
	zAxisAngleOff = 2*PI*frac*current.z;
	zAxisAngle += zAxisAngleOff;
	dist = neo.Length();
	neo.x = cos(zAxisAngle);
	neo.y = sin(zAxisAngle);
	neo = neo * dist;
	final = neo+tem;
	give = final.y;

	return give;

}

//
// getXOffsetFromZ
//	... called by: GTrackArray::getYOffsetFromXYZ
//	... description: for PERP_SPIRAL warp, get the x offset based on the z of the origianl point
//
float GTrackArray::getXOffsetFromZ( float z, float frac, float range )
{
	
	float zAxisAngle = 2*PI*frac*z; // this is the perpspiral equation
	return range*cos(zAxisAngle); // extract the x part

}

//
// getYOffsetFromZ
//	... called by: GTrackArray::getYOffsetFromXYZ
//	... description: for PERP_SPIRAL warp, get the y offset based on the z of the origianl point
//
float GTrackArray::getYOffsetFromZ( float z, float frac, float range )
{
	
	float zAxisAngle = 2*PI*frac*z; // this is the perpsiral equation
	return range*sin(zAxisAngle); // extract the y part

}

//
// getNormal
//	... called by: CMyD3DApplication::RenderUpdateView
//	... description: returns a normal vector depending on what warp we're in.
//
Vec3d GTrackArray::getNormal( float x, float y, float z, float radius, float frac )
{
	if ( warpType != WARP_AXIALSPIRAL )
	{
		Vec3d mine(0,1,0); // any other warp, the normal vector is always up.
		return mine;
	}

	// THIS IS A SLIGHT VARIATION ON THE AXIALSPIRAL WARP CODE
	Vec3d norm;
	Vec3d radial(0,radius,z);
	Vec3d current(x,y,z);
	Vec3d neo = current-radial;
	Vec3d tem = radial-current;
	float zAxisAngle = atan2(neo.y,neo.x);
	float zAxisAngleOff = 2*PI*frac*z;
	zAxisAngle += zAxisAngleOff;
	neo.x = cos(zAxisAngle);
	neo.y = sin(zAxisAngle);
	neo.z = 0;
	Vec3d final = neo+tem;
	Vec3d last = current+final;
	norm.x = radial.x - last.x;
	norm.y = radial.y - last.y;
	norm.z = 0;

	return norm;
}

//
// getXOffsetFromXYZ
//	... called by: any, but mostly CMyD3DApplication::RenderGameObjects()
//	... description: general function, gets x offset depending on what warp we're using
//
float GTrackArray::getXOffsetFromXYZ( float x, float y, float z, float radius, float frac )
{
	switch( warpType )
	{
	case WARP_NONE:
	case WARP_TUNNEL:
		return 0;
	case WARP_CURVY:
		return 0; // doesn't move side to side
	case WARP_AXIALSPIRAL:
		return getAxialXOffset(x,y,z,radius,frac);
	case WARP_PERPSPIRAL:
		return getXOffsetFromZ(z,frac,radius);
	}
	
}

//
// getYOffsetFromXYZ
//	... called by: any, but mostly CMyD3DApplication::RenderGameObjects()
//	... description: general function, gets y offset depending on what warp we're using
//
float GTrackArray::getYOffsetFromXYZ( float x, float y, float z, float radius, float frac )
{
	switch( warpType )
	{
	case WARP_NONE:
	case WARP_TUNNEL:
		return 0;
	case WARP_CURVY:
		return getCurvyYOffset(z,radius,frac);
	case WARP_AXIALSPIRAL:
		return getAxialYOffset(x,y,z,radius,frac);
	case WARP_PERPSPIRAL:
		return getYOffsetFromZ(z,frac,radius);
	}
}

// NOT USED
/*
float GTrackArray::getRotationFromXYZ( float x, float y, float z, float radius, float frac )
{
	float zAxisAngleOff = 0;
	switch ( warpType )
	{
	case WARP_NONE:
	case WARP_TUNNEL:
		break;
	case WARP_PERPSPIRAL:
	case WARP_AXIALSPIRAL:
		zAxisAngleOff = 2*PI*frac*z;
		break;
	}
	

	return zAxisAngleOff;
}
*/

//
// getCurvyYOffset
//	... called by: getYOffsetFromXYZ
//	... description: get CURVY offset
//
float GTrackArray :: getCurvyYOffset( float z, float radius, float frac )
{
	float zAxisAngleOff = 2*PI*frac*z;
	return radius*sin(zAxisAngleOff);
}

